import socket
import time
from bloonix.dbi.classloader import ClassLoader
from bloonix.dbi import DBI
from bloonix.heaven import H
from bloonix import utils

# DONT FORGET TO UPDATE THE MODEL VERSION IN SCHEMA
MODEL_VERSION = 11

DEFAULT_SYS_OPTIONS = [
    ["event_history", "365", "int"],
    ["stat_history", "365", "int"],
    ["stat_hourly_history", "120", "int"],
    ["result_history", "365", "int"],
    ["notification_history", "365", "int"],
    ["user_tracking_history", '365', "int"],
    ["force_email_usernames", "1", "int"],
    ["min_password_length", "8", "int"],
    ["max_password_length", "255", "int"],
    ["password_policy", '["letter"]', "json"],
    ["password_policy_letter", "a-zA-Z", "string"],
    ["password_policy_lowercase", "a-z", "string"],
    ["password_policy_uppercase", "A-Z", "string"],
    ["password_policy_number", '0-9', "string"],
    [
        "password_policy_special",
        '~!@#$%^&*_-+=`|\(){}[]:;"''<>,.?/€',
        "string"
    ],
    [
        "password_policy_special_compiled",
        '~!@#$%\^&*_\-+=`|\\(){}\[\]:;"''<>,.?/€',
        "string"
    ],
    ["plugin_version", "0", "string"],
    ["version", H.version, "string"],
    ["has_maintenance", "0", "int"],
    [
        "server_names",
        H.config.get("bloonix_server_names", socket.gethostname()),
        "string"
    ],
]

class Schema(ClassLoader):
    def __init__(self, config):
        config["logger_name"] = "bloonix-database"
        super().__init__(__path__[0], config)
        db_lock = DBI.setup(**config)
        db_lock.autocommit_default(False)

        while True:
            try:
                db_lock.connect()
                break
            except Exception:
                H.log.exception(level="warning")
                time.sleep(3)

        H.log.warning(
            "waiting for database lock on table "
            "db_manager to upgrade schema")
        db_lock.lock({"db_manager": "write"})
        H.log.warning("table db_manager locked")
        self.run()
        db_lock.commit()
        db_lock.unlock()
        H.log.warning("database schema upgrade done")
        db_lock.disconnect()

    def run(self):
        try:
            self.dbi.connect()
            self.dbi.run_transaction(
                {"sys_options": "write"},
                self.check_sys_options)
            self.dbi.run_transaction(
                {"user": "write", "user_secret": "write"},
                self.check_admin_user)
            self.upgrade_model()
        finally:
            self.dbi.disconnect()

    def check_sys_options(self):
        stmt, bind = self.sql.select(
            table="sys_options",
            columns="*")
        rows = self.dbi.fetchall(stmt, bind)

        options = {}
        for row in rows:
            options[row["key"]] = row["value"]

        for item in DEFAULT_SYS_OPTIONS:
            if item[0] not in options:
                stmt, bind = self.sql.insert(
                    table="sys_options",
                    data={
                        "key": item[0],
                        "value": item[1],
                        "data_type": item[2],
                    })
                self.dbi.do(stmt, bind)

        if "version" in options and options["version"] != H.version:
            stmt, bind = self.sql.update(
                table="sys_options",
                data={"value": H.version},
                condition=["where", ["key", "version"]])
            self.dbi.do(stmt, bind)

    def check_admin_user(self):
        stmt, bind = self.sql.select(
            table=["user", "*"],
            condition=["where", ["id", 1]])
        if not self.dbi.fetchone(stmt, bind):
            stmt, bind = self.sql.insert(
                table="user",
                data={
                    "id": 1,
                    "username": "admin",
                    "name": "Administrator",
                    "locked": 0,
                    "role": "admin",
                    "filter_by_group": 0
                })
            self.dbi.do(stmt, bind)
            password, salt, rounds = utils.pbkdf2_sha512_base64("bloonix")
            stmt, bind = self.sql.insert(
                table="user_secret",
                data={
                    "user_id": 1,
                    "crypt_type": 1,
                    "salt": salt,
                    "rounds": rounds,
                    "password": password,
                })
            self.dbi.do(stmt, bind)

    def upgrade_model(self):
        stmt, bind = self.sql.select(
            table="sys_options",
            columns="value",
            condition=["where", ["key", "model_version"]])
        row = self.dbi.fetchone(stmt, bind)
        model_version = int(row["value"])
        while model_version < MODEL_VERSION:
            model_version += 1
            attr = "upgrade_to_v%s" % model_version
            if hasattr(self, attr):
                func = getattr(self, attr)
                func()
            stmt, bind = self.sql.update(
                table="sys_options",
                data={"value": model_version},
                condition=["where", ["key", "model_version"]])
            self.dbi.do(stmt, bind)

    def upgrade_to_v3(self):
        self.dbi.do("alter table chart_view drop column authkey");

    def upgrade_to_v4(self):
        self.dbi.do(
            "alter table regauth_host add column allow_from "
            "json not null default '[]'")

    def upgrade_to_v5(self):
        self.dbi.do(
            "alter table service_parameter add column satellite_type "
            "smallint not null default '0'")
        self.dbi.do(
            "alter table service_parameter add column satellites "
            "json not null default '[]'")
        for row in self.dbi.fetchall("select * from service_parameter"):
            if ("type" in row["location_options"]
                    and row["location_options"]["type"] == "rotate"):
                stmt, bind = self.sql.update(
                    table="service_parameter",
                    data={
                        "satellite_type": 1,
                        "satellites": row["location_options"]["satellite"]
                    },
                    condition=["where", ["id", row["id"]]])
                self.dbi.do(stmt, bind)
        self.dbi.do("alter table service_parameter drop column location_options")

    def upgrade_to_v6(self):
        if self.config["driver"] == "pg":
            self.dbi.do(
                'alter table "user" add column "style" varchar(20) not null default %s',
                ["default"])
        else:
            self.dbi.do(
                'alter table `user` add column "style" varchar(20) not null default %s',
                ["default"])

    def upgrade_to_v7(self):
        stmt, bind = self.sql.select(
            table="sys_options",
            columns="value",
            condition=["where", ["key", "server_names"]])
        if not self.dbi.fetchone(stmt, bind):
            stmt, bind = self.sql.insert(
                table="sys_options",
                data={
                    "key": "server_names",
                    "value": DEFAULT_SYS_OPTIONS["bloonix_server_names"],
                    "type": "string",
                })
            self.dbi.create(stmt, bind)

    def upgrade_to_v8(self):
        if self.config["driver"] == "pg":
            self.dbi.do(
                'create unlogged table "keepalived" ('
                '    "id" bigserial not null primary key,'
                '    "me" text,'
                '    "time" bigint'
                ')'
            )
        else:
            self.dbi.do(
                'create table `keepalived` ('
                '    `id` bigserial not null primary key,'
                '    `me` text,'
                '    `time` bigint'
                ') engine=innodb'
            )

    def upgrade_to_v9(self):
        if self.config["driver"] == "pg":
            self.dbi.do(
                'create table "user_trusted_browser" ('
                '    "user_id" bigint not null,'
                '    "2fa_sid" varchar(255) not null,'
                '    "expire" bigint not null,'
                '    foreign key ("user_id") references "user"("id") on delete cascade'
                ')'
            )
        else:
            self.dbi.do(
                'create table `user_trusted_browser` ('
                '    `user_id` bigint not null,'
                '    `2fa_sid` varchar(255) not null,'
                '    `expire` bigint not null,'
                '    foreign key (`user_id`) references `user`(`id`) on delete cascade'
                ') engine=innodb'
            )
