from bloonix.dbi.crud import CRUDV
from bloonix.heaven.validator import Validator
from bloonix.heaven import H
from bloonix.json import json
from bloonix import timezone
from bloonix import rx

class User(CRUDV):
    def __init__(self):
        self.has_unique_column = {
            "column": "username",
            "err": "err-618",
        }

    def init(self, validate_admin_account=None):
        sys_options = H.model.db.sys_options.get_keys(
            "min_password_length",
            "max_password_length",
            "force_email_usernames")

        self.validator = Validator(once=True)
        self.validator.set({
            "name": {
                "type": str,
                "maxlength": 50,
                "optional": True
            },
            "phone": {
                "type": str,
                "minlength": 0,
                "maxlength": 30,
                "optional": True,
            },
            "timezone": {
                "type": str,
                "list_of_dicts_key": "value",
                "list_of_dicts_options": timezone.as_form(),
                "default": "Europe/Berlin"
            },
            "language": {
                "type": str,
                "options": ["de", "en"],
                "default": "en"
            },
            "password": {
                "type": str,
                "minlength": sys_options["min_password_length"],
                "maxlength": sys_options["max_password_length"],
                "function": H.plugin.password_policy.check_characters,
                "password": True
            },
            "comment": {
                "type": str,
                "minlength": 0,
                "maxlength": 200,
                "optional": True
            },
            "filter_by_group": {
                "type": int,
                "options": [0, 1],
                "default": 1,
            },
            "2fa_addr": {
                "type": str,
                "minlength": 0,
                "maxlength": 100,
                "optional": True
            },
        })

        if not validate_admin_account:
            self.validator.set({
                "force_password_change": {
                    "type": int,
                    "options": [0, 1],
                    "default": 0,
                },
                "locked": {
                    "type": int,
                    "options": [0, 1],
                    "default": 0,
                },
                "role": {
                    "type": str,
                    "options": ["admin", "operator", "user"],
                    "default": "user",
                }
            })

        if sys_options["force_email_usernames"] == 1:
            self.validator.set(username={
                "type": str,
                "maxlength": 50,
                "regex": r"^(?:admin|%s)\Z" % rx.EMAIL_BASE
            })
        else:
            self.validator.set(username={
                "type": str,
                "maxlength": 50
            })

    def update_setting(self, user_id, column, key, value):
        self.dbi.run_transaction(
            {"user": "write"},
            self.__update_setting,
            user_id, column, key, value)

    def __update_setting(self, user_id, column, key, value):
        row = self.get(user_id)
        settings = row[column]
        settings[key] = value
        settings = json.dumps(settings)
        self.update_by_id(user_id, {column: settings})

    def create_new_user(self, data, secret_data):
        return self.dbi.run_transaction(
            {
                "user": "write",
                "user_secret": "write",
                "user_dashboard": "write",
            },
            self.__create_new_user,
            data,
            secret_data,
        )

    def __create_new_user(self, data, secret_data):
        user = self.create_and_get(data)
        secret_data["user_id"] = user["id"]
        self.schema.user_secret.create(secret_data)
        self.schema.user_dashboard.create_default_dashboard(user["id"], with_monitoring_screen=True)
        return user

    def block_user(self, user_id):
        stmt, bind = self.sql.delete(
            table="user_session",
            condition=[
                "where", ["user_id", user_id]
            ])
        self.dbi.do(stmt, bind)
        stmt, bind = self.sql.update(
            table="user",
            data={"locked": 1},
            condition=[
                "where", ["id", user_id]
            ])
        self.dbi.do(stmt, bind)

    def increase_2fa_tries(self, user_id):
        return self.dbi.run_transaction(
            {"user": "write"},
            self.__increase_2fa_tries,
            user_id)

    def __increase_2fa_tries(self, user_id):
        stmt, bind = self.sql.select(
            table="user",
            columns="2fa_tries",
            condition=["where", ["id", user_id]])
        row = self.dbi.fetchone(stmt, bind)
        stmt, bind = self.sql.update(
            table="user",
            data={"2fa_tries": row["2fa_tries"] + 1},
            condition=["where", ["id", user_id]])
        self.dbi.do(stmt, bind)

    def reset_tries(self, user_id):
        stmt, bind = self.sql.update(
            table="user",
            data={"2fa_tries": 0},
            condition=["where", ["id", user_id]])
        self.dbi.do(stmt, bind)
            
